/* Program Copyright Nathanael Hoyle.  Please read both README and LICENSE files. */

/** @file cdstatus.c main source file for cdstatus.
  * This file contains all the code to do pretty much anything not related to
  * cddb.  It handles program arguments (and config file parameters), does all
  * I/O with the cd/drive, and creates and writes the output files. It calls
  * routines in cdstatus_cddb.c to handle cddb related tasks when applicable.
  */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <sys/ioctl.h>
#include <linux/cdrom.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <unistd.h>
#include <linux/types.h>

#include "cdstatus.h"
#include "cdstatus_cddb.h"
#include "handle_args.h"
#include "basic_info.h"
#include "read_toc.h"
#include "cd_toc_info.h"
#include "build_output_filename.h"
#include "open_output_file.h"
#include "track_listing.h"
#include "cddb_calc_discid.h"

#ifndef PUTSILENT
#define PUTSILENT(A) if(cdsargs->silent==0) { (void)puts((A)); } 
#endif

/** Horribly overgrown function that does all the real work of reading a cd.
  * This function retrieves info about cd "mode" (type of cd) and handles 
  * almost all the work of extracting audio data for the actual ripping process.
  * Probably should be broken up. 
  * @param[in] cdsargs program options/settings
  * @see cdstatus_args 
  */
static int process_audio(const cdstatus_args * cdsargs);


/** Perform an internal driver reset operation.
  * This routine issues an ioctl() to the optical drive telling it to perform
  * a reset. Exactly how much good this does depends on the drive and what it
  * thinks a reset is, as well as whether anything was wrong (and what). Can
  * be useful to unlock a locked drive tray, etc.  Doesn't really need any
  * info except the path to the device, but currently takes pointer to full
  * program options struct.  Usually requires root priviledges to actually
  * perform reset operation. It is run by specifying --reset command-line
  * option, and may not be used in conjunction with any other options (aside
  * from drive path).
  * @param[in] cdsargs program options/settings
  */
static int doReset(const cdstatus_args * cdsargs);

/** outputs wav header to file.
  * Takes currently open file handle for output file and writes the wav header
  * portion of the file, after which the rest of the output to file is performed
  * elsewhere. 
  * @param[in] readframes The number of frames of audio in the track.  Needed to
  *                   calculate values filled in in header.
  * @param[in] audio_out Pre-opened file * for output file to write to.
  * @return nothing (probably should return success/failure)
  */
static void writeWavHeader(unsigned int, FILE *);

/** main routine which does almost no work, calling other functions to act.
  * Displays basic "banner" type output and calls other functions, based on
  * what program options were set.
  * @param[in] argc the system provided argument count
  * @param[in] argv the system provided argument strings
  * @return 0 on success, some other value otherwise
  */
int main(int argc, char *argv[])
{
	int type;
	cdstatus_args cdsargs;

	/*take care of these up front */
	if (argc == 2)
	{
		if (strcmp(argv[1], "--version")==0)
		{
			printf("\ncdstatus version %s by Nathanael Hoyle\n", VERSION);
			return 0;
		}
		else if (strcmp(argv[1], "--help")==0)
		{
			printHelp();
			return 0;
		}
	}

	(void)puts("\nCD Drive Status and CD Analyzer / Reader by Nathanael Hoyle");
	printf("Version %s. Built at %s on %s.\n\n", VERSION, __TIME__, __DATE__);

	/* set default options and override as specified */
	memset((void *) &cdsargs, 0, (size_t) sizeof(cdsargs));
	handleArgs(argc, argv, &cdsargs);

	if (cdsargs.reset!=0)
	{
		return doReset(&cdsargs);
	}

	type = basicInfo(&cdsargs);
	if ((type == CDS_AUDIO) || (type == CDS_MIXED))
	{
		return process_audio(&cdsargs);
	}

	return EXIT_SUCCESS;
}

static int process_audio(const cdstatus_args * cdsargs)
{
	int drive;
	struct cdrom_read_audio cdra;
	struct cdrom_tochdr cdtochdr;
	cd_toc_info cdtocinfo[100];
	int status;
	unsigned int counter, counter2, readframes, retry;
	FILE *audio_out;
	unsigned int hashes;
	unsigned int oldhashes = 0;
	unsigned int start_track, end_track;
	unsigned int discid;
	char filename[1024];
   
	/* safety initialization */
	filename[0] = '\0';
	memset(cdtocinfo, 0, sizeof(cdtocinfo));
	memset(&cdtochdr, 0, sizeof(cdtochdr));
	memset(&cdra, 0, sizeof(cdra));

	drive = open(cdsargs->drivename, O_RDONLY | O_NONBLOCK);
	if (drive == -1)
	{
		printf("Error opening drive %s.\n", cdsargs->drivename);
		perror(NULL);
		return -1;
	}

	if (readTOC(drive, cdtocinfo, &cdtochdr, cdsargs->silent) != 0)
	{
		(void)puts("Error reading TOC.  Exiting.");
		return -1;
	}

	if (cdsargs->cddb!=0)
	{
		discid = calcDiscId(cdtochdr.cdth_trk1, (const cd_toc_info *) cdtocinfo);
		if (cdsargs->silent==0)
		{
			printf("CDDB disc id: %08x\n", discid);
		}

		/*do cddb query */
		cddb_query(cdtochdr.cdth_trk1, (const cd_toc_info *) cdtocinfo, cdsargs->noMangle, cdsargs->cddb_site, cdsargs->default_first_match);

		if (cdsargs->silent==0)
		{
			/*printf("Track listing found for cd %s by %s:\n",album_name,artist_name); */
			for (counter = 1; counter <= cddb_tracks; ++counter)
			{
				printf("Track %u: ", counter);
				if (trackinfo[counter].artist[0] != '\0')
				{
					printf("%s by %s\n", trackinfo[counter].title, trackinfo[counter].artist);
				}
				else
				{
					printf("%s\n", trackinfo[counter].title);
				}
			}
		}
	}

	if (cdsargs->readtest!=0)
	{
		if (cdsargs->start <= 1)
		{
			start_track = cdtochdr.cdth_trk0;
		}
		else
		{
			start_track = (unsigned int)cdsargs->start;
		}
		if ((cdsargs->stop >= cdtochdr.cdth_trk1) || (cdsargs->stop == 0))
		{
			end_track = cdtochdr.cdth_trk1;
		}
		else
		{
			end_track = (unsigned int)cdsargs->stop;
		}

		if (cdsargs->silent==0)
		{
			if (start_track == end_track)
			{
				printf("\n\nBeginning digital audio extraction of track %u.\n", start_track);
			}
			else
			{
				printf("\n\nBeginning digital audio extraction of tracks %u through %u.\n", start_track, end_track);
			}
		}

		cdra.buf = malloc(CD_FRAMESIZE_RAW * (cdsargs->read_chunk_size));
		if (cdra.buf == NULL)
		{
			perror("Unable to allocate enough memory for buffer to hold cd data (cdra.buf). Try decreasing cdsargs->read_chunk_size and try again");
			return -1;
		}

		for (counter = start_track; counter <= end_track; ++counter)
		{
			hashes = oldhashes = 0;
			if ((cdtocinfo[counter].data!=0) && (cdsargs->silent==0))
			{
				printf("Skipping track %u.  (Data track).\n", counter);
				continue;
			}

			if (cdsargs->silent==0)
			{
				printf("Beginning track %u: %s.  Progress:\n", counter, trackinfo[counter].title);
			}

			/* TODO: Finish changing me over */
			buildOutputFilename(cdsargs, counter, filename);
#ifdef DEBUG
			printf("Returned filename: %s\n", filename);
#endif
			audio_out = openOutputFile(filename,cdsargs->lame, counter,cdsargs->lameopts);
			if(audio_out==NULL)
			{
				(void)puts("Got null output file handle, this shouldn't happen... exiting.");
				exit(EXIT_FAILURE);
			}

			cdra.nframes = cdsargs->read_chunk_size;
			cdra.addr_format = CDROM_MSF;

			/* This calculates the length of the track it needs to rip by subtracting its starting offset from the starting offset of the next track.
			 * Only, in the case of the last track, you have to subtract its offset from that of the leadout track, which is actually stored in track 0.
			 * The difference in offsets, in 'frames' is assigned to readframes.
			 */
			if (counter == cdtochdr.cdth_trk1)
			{
				readframes = cdtocinfo[0].frame_global - cdtocinfo[counter].frame_global;
			}
			else
			{
				readframes = cdtocinfo[counter + 1].frame_global - cdtocinfo[counter].frame_global;
			}

			cdra.addr.msf.minute = cdtocinfo[counter].min;
			cdra.addr.msf.second = cdtocinfo[counter].sec;
			cdra.addr.msf.frame = cdtocinfo[counter].frame_local;

			/* If they have asked for wav file output (the default at this point), go ahead and write the wav file header and come back. */
			if (cdsargs->waveout!=0)
			{
				writeWavHeader(readframes, audio_out);
			}
			/* counter2 represents position of next frame to be read */
			for (counter2 = 0; counter2 < readframes; counter2 += (cdsargs->read_chunk_size))
			{
				if ((unsigned int)((cdsargs->read_chunk_size) + counter2) > readframes)
				{
					cdra.nframes = readframes - counter2;
				}

				status = ioctl(drive, CDROMREADAUDIO, &cdra);
				if (status == -1)
				{
					if (errno == EOPNOTSUPP)
					{
						(void)puts("Error, the drive does not support reading in this mode. You will not be able to use the --readtest (--rip) option with this drive, sorry.");
						free(cdra.buf);
						if(cdsargs->lame!=0)
						{
							pclose(audio_out);
						}
						else if(cdsargs->lame==0)
						{
							if(fclose(audio_out)!=0)
							{
								(void)puts("Error closing output file.");
							}
						}
						exit(EXIT_FAILURE);
					}
					if (cdsargs->silent==0)
					{
						printf("\nError trying to read track %u at ", counter);
						printf("minute: %d, second: %d, frame %d: ", cdra.addr.msf.minute, cdra.addr.msf.second, cdra.addr.msf.frame);
						perror(NULL);
					}
					for (retry = 1; (retry <= (unsigned int) (cdsargs->max_retries)) && (status == -1); retry++)
					{
						PUTSILENT("Retrying...")
						status = ioctl(drive, CDROMREADAUDIO, &cdra);
					}
					if (status == -1)
					{
						if (cdsargs->silent==0)
						{
							printf("All %d retries failed. Expect tiny skip in audio.\n", cdsargs->max_retries);
						}
					}
					else
					{
						if (cdsargs->silent==0)
						{
							printf("Succeeded on retry %u.\n", retry - 1);
						}
					}
					(void)fflush(stdout);
				}
				if (status != -1)
				{
					if(fwrite((const void *) cdra.buf, (size_t) CD_FRAMESIZE_RAW, (size_t) cdra.nframes, audio_out)!=(size_t)cdra.nframes)
					{
						perror("Error writing audio frames to wav file.  Exiting");
						if(cdsargs->lame!=0)
						{
							pclose(audio_out);
						}
						else
						{
							if(fclose(audio_out)!=0)
							{
								perror("Error closing output wav file");
							}
						}
						exit(EXIT_FAILURE);
					}
				}

				cdra.addr.msf.frame += cdra.nframes;
				if (cdra.addr.msf.frame >= CD_FRAMES)
				{
					cdra.addr.msf.second += cdra.addr.msf.frame / CD_FRAMES;
					cdra.addr.msf.frame = cdra.addr.msf.frame % CD_FRAMES;
					if (cdra.addr.msf.second >= 60)
					{
						cdra.addr.msf.minute += cdra.addr.msf.second / 60;
						cdra.addr.msf.second = cdra.addr.msf.second % 60;
					}
				}
				if (cdsargs->silent==0)
				{
					hashes = ((counter2 * 78) / readframes) + 1;
					if (hashes > oldhashes)
					{
						hashes -= oldhashes;
						oldhashes += hashes;
						for (; hashes!=0; --hashes)
						{
							(void)putchar('#');
						}
						(void)fflush(stdout);
					}
				}
			}
			if(cdsargs->lame!=0)
			{
				if(pclose(audio_out))
				{
					perror("Error closing pipe to lame");
				}
			}
			if(cdsargs->lame==0)
			{
				if(fclose(audio_out)!=0)
				{
					perror("Error closing output wav file");
				}
			}
			if (cdsargs->silent==0)
			{
				printf("\nTrack %u completed.\n", counter);
			}
		}
	}

	free(cdra.buf);
	return 0;
}

static int doReset(const cdstatus_args * cdsargs)
{
   /** file handle for optical device to reset */
	int drive;

   /** the status code returned from the ioctl() */
	int status;

	(void)puts("Attempting to reset drive.");
	drive = open(cdsargs->drivename, O_RDONLY | O_NONBLOCK);
	if (drive == -1)
	{
		perror("Error opening drive device");
		return -1;
	}
	status = ioctl(drive, CDROMRESET);
	if (status == -1)
	{
		if (errno == EACCES)
		{
			(void)puts("Permission denied error received.  Note that most systems will require root access to reset a drive.");
		}
		else
		{
			perror(NULL);
		}
		(void)puts("Error on resetting drive.  Assume reset failed.");
		return -1;
	}
	else
	{
		(void)puts("Drive successfully reset.");
	}
	return 0;
}

static void writeWavHeader(unsigned int readframes, FILE * audio_out)
{
	long int chunksize;
	typedef struct _wavHeader
	{
		int32_t RIFF_header;
		int32_t total_size;
		int32_t WAVE;
		int32_t fmt;
		int32_t subchunk_size;
		int16_t audio_format;
		int16_t number_channels;
		int32_t sampling_rate;
		int32_t byte_rate;
		int16_t block_align;
		int16_t bits_per_sample;
	} wavHeader;

	wavHeader wHeader;

	/* "RIFF" */
	wHeader.RIFF_header = 0x46464952;

	chunksize = readframes * CD_FRAMESIZE_RAW;
	wHeader.total_size = (int32_t)(chunksize + sizeof(wavHeader));

	/* "WAVEfmt " */
	wHeader.WAVE = 0x45564157;
	wHeader.fmt = 0x20746D66;

	wHeader.subchunk_size = 16;
	wHeader.audio_format = 1;
	wHeader.number_channels = 2;
	wHeader.sampling_rate = 44100;
	wHeader.byte_rate = 176400;
	wHeader.block_align = 4;
	wHeader.bits_per_sample = 16;

	if(fwrite((const void *) &wHeader, sizeof(wavHeader), (size_t) 1, audio_out)!=1)
	{
		perror("Error writing wav file header");
		if(fclose(audio_out)!=0)
		{
			perror("Error closing output wav file");
		}
		exit(EXIT_FAILURE);
	}
	if(fprintf(audio_out, "data")!=4)
	{
		perror("Error writing data header in wav file");
		if(fclose(audio_out)!=0)
		{
			perror("Error closing output wav file");
		}
		exit(EXIT_FAILURE);
	}
	if(fwrite((const void *) &chunksize, sizeof(long int), (size_t) 1, audio_out)!=1)
	{
		perror("Error writing wav file chunksize header");
		if(fclose(audio_out)!=0)
		{
			perror("Error closing output wav file");
		}
		exit(EXIT_FAILURE);
	}

	return;
}
