#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include "handle_args.h"

static void processConfigFile(cdstatus_args * cdsargs);

static void processCommandLine(int argc, char *argv[], cdstatus_args * cdsargs);

void handleArgs(int argc, char *argv[], cdstatus_args * cdsargs)
{
	/* Program defaults */
	strncpy(cdsargs->outputdir, ".", 255);
	strncpy(cdsargs->basename, "track", 255);
	strncpy(cdsargs->drivename, "/dev/cdrom", 255);
	strncpy(cdsargs->cddb_site, "freedb.freedb.org", 255);
	strncpy(cdsargs->lameopts, "--preset standard", 512);
	cdsargs->waveout = 1;
	cdsargs->read_chunk_size = 1;
	cdsargs->max_retries = 3;
	cdsargs->cddb = 0;
	cdsargs->default_first_match = 0;
	cdsargs->start = 1;
	cdsargs->stop = 99;
	cdsargs->lame = 0;
	cdsargs->readtest = 0;
	cdsargs->format_on = 0;
	cdsargs->format[0]='\0'; /* May be a bad idea... dunno yet */
	cdsargs->noMangle = 0;
	cdsargs->silent = 0;


	/* process the config file first, so that the command line overrides the config
	 * file options */
	processConfigFile(cdsargs);
	processCommandLine(argc, argv, cdsargs);

	return;
}

static void processConfigFile(cdstatus_args * cdsargs)
{
	char *home_path;
	char config_name[255];
	char config_line[1024];
	struct stat config_stat;
	FILE *config_file;
	char *optname;
	char *optvalue;

	config_name[0] = '\0';
	config_line[0] = '\0';

	/* .cdstatus config values */
	home_path=getenv("HOME");
	if (home_path==NULL )
	{
		return;
	}
	else if(strlen(home_path) > 240)
	{
		(void)puts("Environment variable HOME too long, .cdstatus config file values will not be used.");
		return;
	}

	(void)snprintf(config_name, 255, "/%s/.cdstatus",home_path);
#ifdef DEBUG
	printf("Using config file %s\n", config_name);
#endif

	if (stat(config_name, &config_stat) == -1)
	{
		if( errno != ENOENT)
		{
 			perror("Error checking for config file, not using");
		}
		else
		{
			(void)puts("Config file not found, not using.");
		}
		return;
	}
	
	if(!S_ISREG(config_stat.st_mode))
	{
		(void)puts("Location pointed to by $HOME/.cdstatus is not a regular file. Ignoring.");
		return;
	}

	config_file = fopen(config_name, "r");
	if (config_file == NULL)
	{
		perror("Error opening config file");
		(void)puts("Config file settings will not be used.");
		return;
	}

	/* past the point where we don't have a config file */
	while (fgets(config_line, 1024, config_file))
	{
		if ((config_line[0] == ';') || (config_line[0] == '\n') || (config_line[0] == ' ') || (config_line[0] == '#'))
		{
			continue;
		}
		if (strchr(config_line, '\n'))
		{
			*(strchr(config_line, '\n')) = '\0';
		}
		optvalue = strchr(config_line, '=');
		if (optvalue == NULL)
		{
			continue;
		}
		*optvalue = '\0';
		++optvalue;
		optname = config_line;

		if (strlen(optvalue) == 0)
		{
			printf("Warning: empty configuration value in user config file .cdstatus (%s) skipped/ignored.\n", optname);
			continue;
		}

		if (strncmp("drive", optname, strlen("drive"))==0)
		{
			strncpy(cdsargs->drivename, optvalue, 255);
		}
		else if (strncmp("outputdir", optname, strlen("outputdir"))==0)
		{
			strncpy(cdsargs->outputdir, optvalue, 255);
		}
		else if (strncmp("retries", optname, strlen("retries"))==0)
		{
			cdsargs->max_retries = atoi(optvalue);
		}
		else if (strncmp("output", optname, strlen("output"))==0)
		{
			if (strncmp(optvalue, "raw", strlen("raw"))==0)
			{
				cdsargs->readtest = 1;
				cdsargs->waveout = 0;
			}
			else if (strncmp(optvalue, "wav", strlen("wav"))==0)
			{
				cdsargs->readtest = 1;
				cdsargs->waveout = 1;
			}
			else
			{
				printf("Invalid output type '%s' in config file\n", optvalue);
				exit(EXIT_FAILURE);
			}
		}
		else if (strncmp("basename", optname, strlen("basename"))==0)
		{
			strncpy(cdsargs->basename, optvalue, 255);
		}
		else if (strncmp("readchunk", optname, strlen("readchunk"))==0)
		{
			cdsargs->read_chunk_size = atoi(optvalue);
		}
		else if (strncmp("cddb_site", optname, strlen("cddb_site"))==0)
		{
			strncpy(cdsargs->cddb_site, optvalue, 255);
		}
		else if (strncmp("no-cddb", optname, strlen("no-cddb"))==0)
		{
			cdsargs->cddb = 0;
		}
		else if (strncmp("cddb", optname, strlen("cddb"))==0)
		{
			if (strncmp(optvalue, "yes", strlen("yes"))==0)
			{
				cdsargs->cddb = 1;
			}
			else if (strncmp(optvalue, "no", strlen("no"))==0)
			{
				cdsargs->cddb = 0;
			}
			else
			{
				printf("Invalid cddb option value in .cdstatus config file: %s\n", optvalue);
			}
		}
		else if (strncmp("format", optname, strlen("format"))==0)
		{
			strncpy(cdsargs->format, optvalue, 512);
			cdsargs->format_on = 1;
		}
		else if (strncmp("default_first_match", optname, strlen("default_first_match"))==0)
		{
			if (strncmp(optvalue, "yes", strlen("yes"))==0)
			{
				cdsargs->default_first_match = 1;
			}
			else if (strncmp(optvalue, "no", strlen("no"))==0)
			{
				cdsargs->default_first_match = 0;
			}
			else
			{
				printf("Invalid value for default_first_match (%s) in .cdstatus config file.\n", optvalue);
			}
		}
	}
	return;
}

static void processCommandLine(int argc, char *argv[], cdstatus_args * cdsargs)
{
	int arg;

	for (arg = 1; arg < argc; ++arg)
	{
		if ((strcmp(argv[arg], "--version")==0) || (strcmp(argv[arg], "--help")==0))
		{
			printf("Sorry, %s is an exclusive option and may not be used in conjunction with other options.\n", argv[arg]);
			exit(EXIT_FAILURE);
		}
		else if (strcmp(argv[arg], "--reset")==0)
		{
			cdsargs->reset = 1;
		}
		else if ((strcmp(argv[arg], "--readtest")==0) || (strcmp(argv[arg], "--rip")==0))
		{
			cdsargs->readtest = 1;
		}
		else if (strcmp(argv[arg], "--start")==0)
		{
			if (arg + 1 >= argc)
			{
				(void)puts("--start requires a value, specifying which track to start ripping at");
				exit(EXIT_FAILURE);
			}
			cdsargs->start = atoi(argv[++arg]);
		}
		else if (strcmp(argv[arg], "--stop")==0)
		{
			if (arg + 1 >= argc)
			{
				(void)puts("--stop requires a value, specifying which track to stop ripping at");
				exit(EXIT_FAILURE);
			}
			cdsargs->stop = atoi(argv[++arg]);
		}
		else if (strcmp(argv[arg], "--retries")==0)
		{
			if (arg + 1 >= argc)
			{
				(void)puts("--retries requires a value, specifying the maximum number of read retries to perform");
				exit(EXIT_FAILURE);
			}
			cdsargs->max_retries = atoi(argv[++arg]);
		}
		else if (strcmp(argv[arg], "--waveout")==0)
		{
			cdsargs->readtest = 1;
			cdsargs->waveout = 1;
		}
		else if (strcmp(argv[arg], "--rawout")==0)
		{
			cdsargs->readtest = 1;
			cdsargs->waveout = 0;
		}
		else if (strcmp(argv[arg], "--drive")==0)
		{
			if (arg + 1 >= argc)
			{
				(void)puts("--drive requires a value, specifying the path to the drive device you which to use (i.e. /dev/hdc)");
				exit(EXIT_FAILURE);
			}
			strncpy(cdsargs->drivename, argv[++arg], 254);
		}
		else if (strcmp(argv[arg], "--outputdir")==0)
		{
			if (arg + 1 >= argc)
			{
				(void)puts("--outputdir requires a value, specifying the directory to place extracted files within (depending on --format, subdirs may be used");
				exit(EXIT_FAILURE);
			}
			strncpy(cdsargs->outputdir, argv[++arg], 254);
		}
		else if (strcmp(argv[arg], "--basename")==0)
		{
			if (arg + 1 >= argc)
			{
				(void)puts("--basename requires a value, see --help");
				exit(EXIT_FAILURE);
			}
			strncpy(cdsargs->basename, argv[++arg], 254);
		}
		else if (strcmp(argv[arg], "--readchunk")==0)
		{
			if (arg + 1 >= argc)
			{
				(void)puts("--readchunk requires a value, see --help");
				exit(EXIT_FAILURE);
			}
			cdsargs->read_chunk_size = atoi(argv[++arg]);
		}
		else if (strcmp(argv[arg], "--cddb")==0)
		{
			cdsargs->cddb = 1;
		}
		else if (strcmp(argv[arg], "--no-cddb")==0)
		{
			cdsargs->cddb = 0;
		}
		else if (strcmp(argv[arg], "--format")==0)
		{
			if (arg + 1 >= argc)
			{
				(void)puts("--format requires a value, see --help");
				exit(EXIT_FAILURE);
			}
			cdsargs->format_on = 1;
			strncpy(cdsargs->format, argv[++arg], 511);
		}
		else if (strcmp(argv[arg], "--nomangle")==0)
		{
			cdsargs->noMangle = 1;
		}
		else if (strcmp(argv[arg], "--silent")==0)
		{
			cdsargs->silent = 1;
		}
		else if (strcmp(argv[arg], "--lame")==0)
		{
			cdsargs->lame = 1;
		}
		else if(strcmp(argv[arg], "--lameopts")==0)
		{
			if(arg+1 >=argc)
			{
				(void)puts("--lameopts requires an option, namely the options to pass to lame (on top of the ones cdstatus uses), you should enclose the arguments to --lameopts with \" quotes.");
				exit(EXIT_FAILURE);
			}
			strncpy(cdsargs->lameopts, argv[++arg], 512);
		}
		else
		{
			(void)puts("Invalid option selected.  Please verify options.");
			printf("Invalid option passed: %s\n\n", argv[arg]);
			exit(EXIT_FAILURE);
		}
	}

	if (cdsargs->reset!=0)
	{
		if (cdsargs->readtest!=0)
		{
			(void)puts("Error, cannot use data extraction and --reset.");
			exit(EXIT_FAILURE);
		}
	}

	if (cdsargs->max_retries < 0)
	{
		(void)puts("Warning: negative value specified for retry count, setting to 0.");
		cdsargs->max_retries = 0;
	}
	if (cdsargs->read_chunk_size > 75)
	{
		(void)puts("Warning: readchunk value over 75 will not work properly on most hardware.  Dropping to 75.");
		cdsargs->read_chunk_size = 75;
	}
	else if (cdsargs->read_chunk_size <= 0)
	{
		(void)puts("Warning: readchunk value of 0 or below will cause all kinds of problems.  Fixing to 1.");
		cdsargs->read_chunk_size = 1;
	}

	if (cdsargs->stop < cdsargs->start)
	{
		(void)puts("--stop value must be greater than or equal to --start value (which defaults to 1)");
		exit(EXIT_FAILURE);
	}
	if (cdsargs->start < 1)
	{
		(void)puts("--start value must be greater than or equal to 1");
		exit(EXIT_FAILURE);
	}

	return;
}
